#include "buffer.h"
#include "draw.h"
#include "3D.h"

#define CUBE_CHUNK_X 10
#define CUBE_CHUNK_Y 10
#define CUBE_CHUNK_Z 5




//GET A CUBE DEFINED BY ITS POS

//cubeChunk is a "cubemap" array
//x,y,z are coordinates
int getCube(int *cubeChunk,double x,double y,double z)
{
  int index=-2,temp=0;
  if(x/2.0>0 && y/2.0>0 && (z-2.0)/2.0>0 && x/2.0<CUBE_CHUNK_X && y/2.0<CUBE_CHUNK_Y && (z-2.0)/2.0<CUBE_CHUNK_Z)
  {
    x=x/2.0;
    y=y/2.0;
    z=(z-2.0)/2.0;
    temp=(int)z;
    index=temp*CUBE_CHUNK_X*CUBE_CHUNK_Y;
    temp=(int)x;
    index+=temp*CUBE_CHUNK_Y;
    temp=(int)y;
    index+=temp;/*
    index=(z-2.0)/2.0*CUBE_CHUNK_X*CUBE_CHUNK_Y+x/2.0*CUBE_CHUNK_Y+y/2.0;
    printf("%d\n",index);*/
  }
  if(index>=0 && index<CUBE_CHUNK_Z*CUBE_CHUNK_X*CUBE_CHUNK_Y)
    return cubeChunk[index];
  else
    return 1;
}


//HEAP SORT UTIL -- CALLED BY sortCubeList()

//this function is a modified version of wikipedia's heap sort
//thanks wikipedia :)

void heap(double *dist,int *index,int i,int n)
{
  int k=i;
  int j=2*k;
  while(j<=n)
  {
    if(j<n && dist[j]>dist[j+1])
    {
      j++;
    }
    if(dist[k]>dist[j])
    {
      invertTableDoubles(dist,k,j);
      invertTableElems(index,k,j);
      k=j;
      j=2*k;
    }
    else
    {
      break;
    }
  }
}


//SORT ALL THE CUBES WHICH WILL BE DISPLAYED

//cubeList is an array which contains the coordinates of the center of each cube
//colorList is the array which contains the color of each cube
//size is the number of cubes

void sortCubeList(double *cubeList,int *colorList,int size)
{
  int i=0;
  double *dist=malloc(size*sizeof(double));
  int *index=malloc(size*sizeof(int));
  
  //calculates distance of each cube
  for(i=0;i<size;i++)
  {
    dist[i]=cubeList[i*3]*cubeList[i*3]+cubeList[i*3+1]*cubeList[i*3+1]+cubeList[i*3+2]*cubeList[i*3+2];
    index[i]=i;
  }
  
  //heap sort
  for(i=size/2-1;i>=0;i--)
  {
    heap(dist,index,i,size-1);
  }
  for(i=size-1;i>=1;i--)
  {
    invertTableDoubles(dist,i,0);
    invertTableElems(index,i,0);
    heap(dist,index,0,i-1);
  }
  
  //sort cubes with given indexes
  double *cubeListImg=malloc(size*3*sizeof(double));
  memcpy(cubeListImg,cubeList,size*3*sizeof(double));
  
  int *colorListImg=malloc(size*sizeof(int));
  memcpy(colorListImg,colorList,size*sizeof(int));
  
  for(i=0;i<size;i++)
  {
    cubeList[i*3]=cubeListImg[index[i]*3];
    cubeList[i*3+1]=cubeListImg[index[i]*3+1];
    cubeList[i*3+2]=cubeListImg[index[i]*3+2];
    colorList[i]=colorListImg[index[i]];
  }
  
  free(dist);
  free(cubeListImg);
  free(colorListImg);
}




//COMPUTE CUBES POS AND CUBES WHICH WILL BE DISPLAYED

//chunk is a "cubemap" array
//cubeList is the array where the pos of each cube will be stored
//colorList is the array where the color of each cube will be stored
//size is the place where the number of displayed cubes will be stored
//x and z rotation is the camera

void computeCubeWorldChunk(int *chunk,double *cubeList,int *colorList,int *size,double xRotation,double zRotation,pos playerPos)
{
  int i=0,j=0,k=0; //iterators
  int index=0,n=0,displayed=0; //utils
  
  //loop to see all cubes
  for(i=0;i<CUBE_CHUNK_Z;i++)
  {
    for(j=0;j<CUBE_CHUNK_X;j++)
    {
      for(k=0;k<CUBE_CHUNK_Y;k++)
      {
        //optimization of cubes display
        index=i*CUBE_CHUNK_Y*CUBE_CHUNK_X+j*CUBE_CHUNK_Y+k;
        displayed=1;
        
        if(i>0 && j>0 && k>0 && i<CUBE_CHUNK_Z-1 && j<CUBE_CHUNK_X-1 && k<CUBE_CHUNK_Y-1)
        {
          displayed=0;
          if(chunk[index+1]==0 || chunk[index-1]==0 || 
            chunk[index+CUBE_CHUNK_Y]==0 || chunk[index-CUBE_CHUNK_Y]==0 || 
            chunk[index+CUBE_CHUNK_X*CUBE_CHUNK_Y]==0 || chunk[index-CUBE_CHUNK_X*CUBE_CHUNK_Y]==0)
            displayed=1;
        }
        
        if(chunk[index]!=0 && displayed==1)
        {
          cubeList[n*3]=(double)j*2.0+1.0;
          cubeList[n*3+1]=(double)k*2.0+1.0;
          cubeList[n*3+2]=(double)i*2.0+1.0;
          colorList[n]=chunk[index];
          n=n+1;
        }
      }
    }
  }
  translate(cubeList,n,-playerPos.x,-playerPos.y,-playerPos.z);
  rotateZ(cubeList,n,zRotation);
  rotateX(cubeList,n,xRotation);
  sortCubeList(cubeList,colorList,n);
  *size=n;
}



//DRAW COMPUTED CUBES

//cubeList is an array which contains coordinates of the center of all the cubes {x1,y1,z1, x2,y2,z2,...}
//colorList is an array which contains the color of each cube
//size is the number of cubes
//x and z rotation is the camera

void drawCubeList(BUF *buffer,double *cubeList,int *colorList,int size,double xRotation,double zRotation)
{
  //one cube -- to translate to each pos of cubeList
  double cubeVertices[8*3]={1.0,1.0,1.0,  1.0,1.0,-1.0,  1.0,-1.0,1.0,  1.0,-1.0,-1.0,  -1.0,1.0,1.0,  -1.0,1.0,-1.0,  -1.0,-1.0,1.0,  -1.0,-1.0,-1.0};
  double cubeVerticesImg[8*3]={0.0};
  
  int cubeFaces[6*4]={1,2,4,3,  1,5,6,2,  1,3,7,5,  8,7,3,4,  8,7,5,6,  8,4,2,6};
  int colors[6]={2,4,6,8,10,12};
  int verticesPos[8*2]={0};
  int i=0;
  
  rotateZ(cubeVertices,8,zRotation);
  rotateX(cubeVertices,8,xRotation);
  
  //draw all cubes
  for(i=0;i<size;i++)
  {
    memcpy(cubeVerticesImg,cubeVertices,24*sizeof(double));
    
    translate(cubeVerticesImg,8,cubeList[i*3],cubeList[i*3+1],cubeList[i*3+2]);
    
    colors[0]=colorList[i];
    colors[1]=colors[0]+1;
    colors[2]=colors[0]-1;
    colors[3]=colors[0]+1;
    colors[4]=colors[0];
    colors[5]=colors[0]-1;
    
    computeVertices(cubeVerticesImg,verticesPos,8);
    if(verticesPos[0]!=-10000)
      drawCube(buffer,cubeVerticesImg,verticesPos,cubeFaces,colors);
  }
}



void initWorld(int *world)
{
  int i=0;
  for(i=0;i<CUBE_CHUNK_X*CUBE_CHUNK_Y*CUBE_CHUNK_Z;i++)
  {world[i]=0;}
  for(i=0;i<CUBE_CHUNK_X*CUBE_CHUNK_Y;i++)
  {
    world[i]=8;
  }
  world[199]=3;
  world[198]=3;
  world[197]=3;
  world[189]=3;
  world[179]=3;
  world[178]=3;
  world[177]=3;
  world[299]=3;
  world[298]=3;
  world[297]=3;
  world[289]=3;
  world[279]=3;
  world[278]=3;
  world[277]=3;
  world[399]=3;
  world[398]=3;
  world[397]=3;
  world[389]=3;
  world[388]=3;
  world[387]=3;
  world[379]=3;
  world[378]=3;
  world[377]=3;
}



////////////////////////
///////   MAIN   ///////
////////////////////////
///(yes, I need this)///


int main(void)
{
  //vars
  BUF *buffer=malloc(SCR_SIZE);
  uint16_t time=0,delay=0; //timers
  char exit=0;
  double angleX=5.0,angleZ=45.0; //camera rotation
  int *world=malloc(CUBE_CHUNK_X*CUBE_CHUNK_Y*CUBE_CHUNK_Z*sizeof(int)); //cube world (kind of a 3D array)
  int *colorList=malloc(CUBE_CHUNK_X*CUBE_CHUNK_Y*CUBE_CHUNK_Z*sizeof(int)); //color of all displayed cubes
  int size=0; //number of displayed cubes
  double *cubeList=malloc(CUBE_CHUNK_X*CUBE_CHUNK_Y*CUBE_CHUNK_Z*3*sizeof(double)); //a chunk of the world to display
  
  pos playerPos;
  playerPos.x=5.0;
  playerPos.y=5.0;
  playerPos.z=4.1;
  
  
  //initialization
  initWorld(world);
  
  bufFill(buffer,0xFF);
  bufDisplay(buffer);
  
  //render loop
  while(!exit)
  {
    while(!any_key_pressed())
    {
      idle();
    }
    //timers update
    time=*(volatile uint16_t*)0x900D000C;
    
    //key detection
    if(isKeyPressed(KEY_NSPIRE_UP))
      angleX-=5.0;
    if(isKeyPressed(KEY_NSPIRE_DOWN))
      angleX+=5.0;
    if(isKeyPressed(KEY_NSPIRE_RIGHT))
      angleZ+=5.0;
    if(isKeyPressed(KEY_NSPIRE_LEFT))
      angleZ-=5.0;
    if(isKeyPressed(KEY_NSPIRE_A)) //forward
    {
      if(getCube(world,playerPos.x+cos(angleZ-90.0),playerPos.y,playerPos.z)==0)
        playerPos.x+=cos(angleZ-90.0);
      if(getCube(world,playerPos.x,playerPos.y+sin(angleZ+90.0),playerPos.z)==0)
        playerPos.y+=sin(angleZ+90.0);
    }
    if(isKeyPressed(KEY_NSPIRE_Z)) //backward
    {
      if(getCube(world,playerPos.x-cos(angleZ-90.0),playerPos.y,playerPos.z)==0)
        playerPos.x-=cos(angleZ-90.0);
      if(getCube(world,playerPos.x,playerPos.y-sin(angleZ+90.0),playerPos.z)==0)
        playerPos.y-=sin(angleZ+90.0);
    }
    if(isKeyPressed(KEY_NSPIRE_ESC))
      exit=1;
      
    //draw
    bufFill(buffer,0xFF);
    computeCubeWorldChunk(world,cubeList,colorList,&size,angleX,angleZ,playerPos);
    drawCubeList(buffer,cubeList,colorList,size,angleX,angleZ);
    bufDisplay(buffer);
    
    //delay if too fast
    delay=*(volatile uint16_t*)0x900D000C-time;
    while(delay<3000)
    {
      idle();
      delay=*(volatile uint16_t*)0x900D000C-time;
    }
  }
  
  free(buffer);
  free(cubeList);
  free(world);
  free(colorList);
  return 0;
}
